import json
import logging
from abc import abstractmethod
from typing import Iterable, Optional

import pandas as pd
import pyarrow as pa

from odp.tabular_v2.util.exp import BindVar, BinOp, Field, Op, Parens, Scalar, UnaryOp

SMALL_MAX = 256
STR_LIMIT = 128  # when to start using a reference
STR_MIN = 12  # what to keep as prefix in the reference
MAX_BIGFILE_SIZE = 64 * 1024 * 1024  # max size of a big file


def convert_schema_outward(schema: pa.Schema) -> pa.Schema:
    """drops the .ref fields"""
    out = []
    for name in schema.names:
        field: pa.Field = schema.field(name)
        if name.endswith(".ref") and name[:-4] in schema.names and field.type == pa.string():
            continue  # skip
        out.append(field)
    return pa.schema(out)


class BigCol:
    def __init__(self):
        pass

    @abstractmethod
    def fetch(self, md5: str) -> bytes:
        """fetch data, called often, should cache"""
        raise NotImplementedError()

    @abstractmethod
    def upload(self, md5: str, data: Iterable[bytes]):
        """upload data"""
        raise NotImplementedError()

    def decode(self, df: pd.DataFrame) -> pd.DataFrame:
        cache = {}  # FIXME: can this use too much memory?

        refs = []
        for name in df.columns:
            if name.endswith(".ref"):
                refs.append(name)

        if not refs:
            return df

        def decode_by_row(row):
            for name in refs:
                ref = row[name]
                if not ref:
                    continue

                target = name[:-4]
                big_id, start, size = ref.split(":")
                start = int(start)
                size = int(size)
                if big_id in cache:
                    data = cache[big_id]
                else:
                    data = self.fetch(big_id)
                    cache[big_id] = data
                if isinstance(row[target], str):  # the field must contain the prefix, from which we infer the type
                    row[target] = data[start : start + size].decode("utf-8")
                else:
                    row[target] = data[start : start + size]
            return row

        df = df.apply(decode_by_row, axis=1)
        df.drop(columns=refs, inplace=True)
        return df


def inner_exp(schema: pa.Schema, op: Optional[Op]) -> Optional[Op]:
    if op is None:
        return None

    fields = []
    for name in schema.names:
        field: pa.Field = schema.field(name)
        if field.type != pa.string() and field.type != pa.binary():
            continue
        if field.metadata and b"big" in field.metadata:
            fields.append(name)

    # TODO don't use the visitor, instead parse manually and use negation context
    def visitor(neg: bool, op: Op) -> Op:
        if isinstance(op, Field):
            return op
        if isinstance(op, Scalar):
            return op
        if isinstance(op, BindVar):
            return op
        if isinstance(op, Parens):
            op.exp = visitor(neg, op.exp)
            return op
        if isinstance(op, UnaryOp):
            if op.prefix in ["~", "not", "!", "invert"]:
                return UnaryOp(prefix=op.prefix, exp=visitor(~neg, op.exp), suffix=op.suffix)
            return op
        if isinstance(op, BinOp):
            op = BinOp(left=visitor(neg, op.left), op=op.op, right=visitor(neg, op.right))
            if isinstance(op.left, Field):
                if str(op.left) in fields:
                    return _inner_exp_binop(neg, op.left, op.op, op.right)
                return op
            elif isinstance(op.right, Field):
                try:
                    op = op.flip()
                except NotImplementedError:
                    logging.warning("can't flip big-col expression: %s", op)
                    return Scalar(src="True", type="bool")
                return visitor(neg, op)
            else:
                return op
        raise ValueError(f"can't convert big-col expression: {type(op)}")

    op = visitor(False, op)
    logging.info("big: inner_exp: %s", repr(op))
    return op


def _inner_exp_binop_str(neg: bool, field: Field, op: str, right: str) -> Op:
    if len(right) > STR_MIN:
        a = right[:STR_MIN]
        b = right[: STR_MIN - 1] + chr(ord(right[STR_MIN - 1]) + 1)
        logging.info("big: str: %s .. %s", json.dumps(a), json.dumps(b))

        if op == "==":
            if neg:
                return Scalar.from_py(False)
            return BinOp(
                left=BinOp(
                    left=Scalar.from_py(a),
                    op="<",
                    right=field,
                ),
                op="and",
                right=BinOp(
                    left=field,
                    op="<",
                    right=Scalar.from_py(b),
                ),
            )
        elif op == "!=":
            if neg:
                return Scalar.from_py(False)
            else:
                return Scalar.from_py(True)
        elif op == ">" or op == ">=":
            return BinOp(
                left=field,
                op=op,
                right=Scalar.from_py(a),
            )
        elif op == "<" or op == "<=":
            return BinOp(
                left=field,
                op=op,
                right=Scalar.from_py(b),
            )
    else:
        return BinOp(
            left=field,
            op=op,
            right=Scalar.from_py(right),
        )
    logging.error("can't convert big-col expression: %s %s %s", field, op, right)
    raise ValueError("can't convert big-col expression")


def _inner_exp_binop(neg: bool, left: Field, op: str, right: Op) -> Op:
    if isinstance(right, Scalar):
        v = right.to_py()
        if isinstance(v, str):
            return _inner_exp_binop_str(neg, left, op, v)
        else:
            raise ValueError("can't convert big-col expression for scalar %s", right)
    raise ValueError("can't convert big-col expression: %s %s %s", left, op, right)



import io
import logging
from typing import Dict, Iterator, List, Union

import pandas as pd
import pyarrow as pa
from pyarrow.lib import ArrowInvalid

from odp.tabular_v2.client import Table
from odp.tabular_v2.client.validation import validate_data_against_schema
from odp.tabular_v2.util import exp


class Transaction:
    """
    a transaction is created implicitly when a table is used as a context manager:

        with table as tx:
            ...

    transaction should be used to modify the data, and make the modifications atomic (which means users won't see
    the changes while they are being made, but only all at once when the transaction is committed at the end).

    they transaction won't commit (and instead rollback) if an exception is raised inside the block.

    when a transaction is created, it might buffer some data locally to improve the performance of the system.
    """

    def __init__(self, table: Table, tx_id: str):
        if not tx_id:
            raise ValueError("tx_id must not be empty")
        self._table = table
        self._id = tx_id
        self._buf: List[Union[pa.RecordBatch, List[Dict]]] = []
        self._buf_rows = 0
        self._buf_size = 0

    # FIXME(oha) this is broken, since we have data buffered locally,
    # we could just flush the data first and make this work
    # or instead put the effort to use the local buffer as false positives
    # since there is no strong use case, this can be done later when the rest is sorted
    # def select(self, query: Union[exp.Op, str, None] = None) -> Iterator[Dict]:
    #    for row in self._table.select(query).rows():
    #        yield row

    def replace(self, query: Union[exp.Op, str, None] = None, vars: Union[Dict, List, None] = None) -> Iterator[Dict]:
        """perform a two-step replace:
        rows that don't match the query are kept.
        rows that match are removed and sent to the caller.
        the caller might insert them again or do something else.

        NOTE: internally, the server might have to send false positives (because of bigcol), which means
        the SDK will have to check for them and insert them back into the table.
        This happens internally and is not exposed to the user.
        """
        if query is None:
            raise ValueError("For your own safety, please provide a query like 1==1")
        assert self._buf_rows == 0  # FIXME: handle buffered data in replace/select
        for row in self._table._query_cursor(type="replace", query=query, vars=vars, tx_id=self._id).rows():
            yield row

    def delete(self, query: Union[exp.Op, str, None] = None) -> int:
        """
        delete rows that match the query

        Note: similarly to the replace, some rows might be false positive and should be added back, but this
        happens internally and is not exposed to the user.
        Returns how many rows were changed
        """
        ct = 0
        for _ in self.replace(query):  # Note(oha) we must iterate over the generator to make it work
            ct += 1
        return ct

    def flush(self):
        """
        flush the data to the server, in case some data is buffered locally
        """
        logging.info("flushing to stage %s", self._id)
        if len(self._buf) == 0:
            return

        schema = self._table.schema()

        buf = io.BytesIO()
        w = pa.ipc.RecordBatchStreamWriter(buf, schema)

        for b in self._buf:
            if isinstance(b, list):
                b = pa.RecordBatch.from_pylist(b, schema=schema)

            try:
                w.write_batch(b)
            except ArrowInvalid as e:
                raise ValueError("Invalid arrow format") from e
        w.close()
        self._table._client._request(
            path="/api/table/v2/sdk/insert",
            params={
                "table_id": self._table._id,
                "tx_id": self._id,
            },
            data=buf.getvalue(),
        ).json()
        self._buf = []
        self._buf_rows = 0

    def insert(self, data: Union[Dict, List[Dict], pa.RecordBatch, pd.DataFrame]):
        """
        add data to the internal buffer to be inserted into the table
        if the buffered data is enough, it will be automatically flushed

        accept a single dictionary, a list of dictionaries, a pandas DataFrame, or a pyarrow RecordBatch
        """
        if isinstance(data, dict):
            data = [data]

        schema = self._table.schema()

        validate_data_against_schema(data, schema)
        if isinstance(data, list):
            # we expand the last list if it's already a list
            last = self._buf[-1] if self._buf else None
            if last and isinstance(last, list):
                last.extend(data)
            else:
                self._buf.append(data)
            self._buf_rows += len(data)
        elif isinstance(data, pd.DataFrame):
            data = pa.RecordBatch.from_pandas(data, schema=schema)
            self._buf.append(data)
            self._buf_rows += data.num_rows
            self._buf_size += data.nbytes
        elif isinstance(data, pa.RecordBatch):
            self._buf.append(data)
            self._buf_rows += data.num_rows
            self._buf_size += data.nbytes
        else:
            raise ValueError(f"unexpected type {type(data)}")

        if self._buf_rows >= 10_000 or self._buf_size >= 10_000_000:
            self.flush()



import copy
import logging
from typing import Optional

import shapely

from odp.tabular_v2.util import exp


class _QueryContext:
    def __init__(self, geo_fields: list[str]):
        self.geo_fields = geo_fields
        self.negate = False

    def neg(self) -> "_QueryContext":
        c = copy.copy(self)
        c.negate = ~c.negate
        return c

    def is_geo_field(self, op: Optional[exp.Op]) -> bool:
        if isinstance(op, exp.Field):
            return op.name in self.geo_fields
        return False

    def convert(self, op: Optional[exp.Op]) -> Optional[exp.Op]:
        if op is None:
            return None
        if isinstance(op, exp.Parens):
            inner = self.convert(op.exp)
            if isinstance(inner, exp.Parens):
                return inner
            return exp.Parens(inner)
        if isinstance(op, exp.BinOp):
            if self.is_geo_field(op.left) or self.is_geo_field(op.right):
                if op.op in ["intersect", "intersects", "contain", "contains", "within", "==", "~"]:
                    return self._convert_intersect(op)
            elif op.op in ["intersect", "intersects", "contain", "contains", "within", "~"]:
                raise ValueError(f"can't do '{op.op}' on non-geo fields")
            left = self.convert(op.left)
            right = self.convert(op.right)
            return exp.BinOp(left=left, op=op.op, right=right)
        if isinstance(op, exp.Field):
            return op
        if isinstance(op, exp.Scalar):
            return op
        if isinstance(op, exp.UnaryOp):
            cur = self
            if op.prefix == "~":
                cur = self.neg()
            return exp.UnaryOp(prefix=op.prefix, exp=cur.convert(op.exp), suffix=op.suffix)
        if isinstance(op, exp.Func):
            cur = self
            if op.name == "invert":
                cur = self.neg()
            args = [cur.convert(a) for a in op.args]
            return exp.Func(name=op.name, args=args)
        if isinstance(op, exp.BindVar):
            return op
        raise ValueError(f"can't convert {op}: {type(op)}")

    def _convert_intersect(self, op: exp.BinOp) -> exp.Op:
        if isinstance(op.left, exp.Field):
            if isinstance(op.right, exp.Scalar):
                val = op.right.to_py()
                try:
                    geo = shapely.from_wkt(val)
                except Exception as e:
                    raise ValueError(f"can't convert '{val}' to geometry: {e}") from e
                return self._intersect_field(op.left, geo)
            # if isinstance(op.right, exp.Field):
            # return exp.Scalar.from_py(~self.negate)

        if isinstance(op.right, exp.Field):
            if isinstance(op.left, exp.Scalar):
                geo = shapely.from_wkt(op.left.to_py())
                return self._intersect_field(op.right, geo)

        raise ValueError(f"unsupported: {type(op.left)} {op.op} {type(op.right)}")

    def _intersect_field(self, field: exp.Field, geo: shapely.Geometry) -> exp.Op:
        logging.info("intersecting field '%s' with '%s'", field, geo)
        fx = exp.Field(name=field.name + ".x")
        fy = exp.Field(name=field.name + ".y")
        fq = exp.Field(name=field.name + ".q")
        x0, y0, x1, y1 = shapely.bounds(geo).tolist()
        if self.negate:
            xop = exp.BinOp(
                exp.Parens(exp.BinOp(exp.BinOp(fx, "-", fq), ">=", exp.Scalar.from_py(x0))),
                "and",
                exp.Parens(exp.BinOp(exp.BinOp(fx, "+", fq), "<=", exp.Scalar.from_py(x1))),
            )
            yop = exp.BinOp(
                exp.Parens(exp.BinOp(exp.BinOp(fy, "-", fq), ">=", exp.Scalar.from_py(y0))),
                "and",
                exp.Parens(exp.BinOp(exp.BinOp(fy, "+", fq), "<=", exp.Scalar.from_py(y1))),
            )
        else:
            xop = exp.BinOp(
                exp.Parens(exp.BinOp(exp.BinOp(fx, "+", fq), ">=", exp.Scalar.from_py(x0))),
                "and",
                exp.Parens(exp.BinOp(exp.BinOp(fx, "-", fq), "<=", exp.Scalar.from_py(x1))),
            )
            yop = exp.BinOp(
                exp.Parens(exp.BinOp(exp.BinOp(fy, "+", fq), ">=", exp.Scalar.from_py(y0))),
                "and",
                exp.Parens(exp.BinOp(exp.BinOp(fy, "-", fq), "<=", exp.Scalar.from_py(y1))),
            )
        return exp.Parens(exp.BinOp(xop, "and", yop))


def test_query():
    op = exp.parse("color == 'red' and not (area intersect 'POLYGON ((0 0, 0 1, 1 1, 1 0, 0 0))')")
    logging.info("'%s'...", op)
    c = _QueryContext(["area"])
    op2 = c.convert(op)
    logging.info("'%s'...", op2)
    assert "color == 'red'" in str(op2)
    assert "area.x - area.q >= 0" in str(op2)  # inverted sign

    # check that raises exception if intersect with no geo field
    try:
        op = exp.parse("other_field intersect 'POLYGON ((0 0, 0 1, 1 1, 1 0, 0 0))'")
        c = _QueryContext([])
        c.convert(op)
    except ValueError as e:
        assert "intersect" in str(e)
    else:
        assert False



from concurrent.futures import Future
from concurrent.futures.thread import ThreadPoolExecutor

pool = ThreadPoolExecutor(4)  # this is a mix of CPU and waiting for CDF...


# TODO(oha): can we remove this?
class Group:
    """
    Create a context of futures that will be awaited on exit
    this prevents the risk of missing an exception in a future
    """

    def __init__(self):
        self.futures = []

    def submit(self, fn, *args, **kwargs) -> Future:
        """submit a new task to the pool"""
        f = pool.submit(fn, *args, **kwargs)
        self.futures.append(f)
        return f

    def results(self) -> list:
        """early return the results of all futures submitted so far"""
        out = [f.result() for f in self.futures]
        self.futures = []
        return out

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.results()



import logging
from typing import Dict, Union

from odp.tabular_v2.util import exp
from odp.tabular_v2.util.util import FieldNotFound


class MinMax:
    def __init__(self, *vals):
        self.min = None
        self.max = None
        self.nulls = False
        for v in vals:
            self.set(v)

    def set(self, v):
        if v is None:
            self.nulls = True
        elif self.min is None:
            self.min = v
            self.max = v
        elif v is True:
            self.max = True
        elif v is False:
            self.min = False
        else:
            self.min = min(self.min, v)
            self.max = max(self.max, v)
        return self

    def __str__(self):
        return f"mm({self.min}, {self.max}, {self.nulls})"

    def can_be_true(self):
        if self.min is None:
            return False
        if isinstance(self.min, bool):
            return self.max
        if isinstance(self.min, Union[int, float]):
            return self.min != 0 or self.max != 0
        raise NotImplementedError(f"can_be_true not implemented for {self}")

    def can_be_false(self):
        if self.min is None:
            return False
        if isinstance(self.min, bool):
            return self.min is False
        if isinstance(self.min, Union[int, float]):
            if self.min == 0:
                return self.max > 0
            elif self.min < 0:
                return self.max >= 0
            else:
                return False
        raise NotImplementedError(f"can_be_true not implemented for {self}")

    def bin_op(self, op: str, other: "MinMax"):
        out = MinMax()
        out.nulls = self.nulls or other.nulls
        if self.min is None:
            out.min = other.min
            out.max = other.max
            return out
        if other.min is None:
            out.min = self.min
            out.max = self.max
            return out

        def reduce(op):
            for left in self.min, self.max:
                for right in other.min, other.max:
                    try:
                        out.set(op(left, right))
                    except ArithmeticError as e:
                        logging.warning("ArithmeticError: %s %s %s: %s", left, op, right, e)
                        out.set(None)
            return out

        if op == "and":
            if self.can_be_true() and other.can_be_true():
                out.set(True)
            if self.can_be_false() or other.can_be_false():
                out.set(False)
            return out
        if op == "or":
            if self.can_be_true() or other.can_be_true():
                out.set(True)
            if self.can_be_false() and other.can_be_false():
                out.set(False)
            return out

        if op == "+":
            return out.set(self.min + other.min).set(self.max + other.max)
        if op == "-":
            return out.set(self.min - other.max).set(self.max - other.min)
        if op == "*":
            return reduce(lambda a, b: a * b)
        if op == "/":
            return reduce(lambda a, b: a / b)

        if op == "<":
            return reduce(lambda a, b: a < b)
        if op == "<=":
            return reduce(lambda a, b: a <= b)
        if op == ">":
            return reduce(lambda a, b: a > b)
        if op == ">=":
            return reduce(lambda a, b: a >= b)
        if op == "==":
            if self.min <= other.max and self.max >= other.min:
                out.set(True)
            if self.min != self.max or self.min != other.min or self.min != other.max:
                out.set(False)
            return out
        if op == "!=":
            return self.bin_op("==", other).unary_op("not")

        raise NotImplementedError(f"bin op {op} not supported")

    def unary_op(self, op: str):
        if op in ["not", "!", "~"]:
            out = MinMax()
            out.nulls = self.nulls
            if self.can_be_true():
                out.set(False)
            if self.can_be_false():
                out.set(True)
            return out
        raise NotImplementedError(f"unary op {op} not supported")


def minmax(op: exp.Op, row: Dict[str, MinMax]) -> MinMax:
    if isinstance(op, exp.BinOp):
        left = minmax(op.left, row)
        right = minmax(op.right, row)
        return left.bin_op(op.op, right)
    if isinstance(op, exp.UnaryOp):
        return minmax(op.exp, row).unary_op(op.prefix)
    if isinstance(op, exp.Func):
        raise NotImplementedError(f"func {op.name} not supported")
    if isinstance(op, exp.Field):
        try:
            return row[str(op)]
        except KeyError:
            raise FieldNotFound(str(op))
    if isinstance(op, exp.Scalar):
        return MinMax(op.to_py())
    if isinstance(op, exp.Parens):
        return minmax(op.exp, row)
    if isinstance(op, exp.BindVar):
        return MinMax(op.val)
    raise ValueError(f"unexpected op {type(op)}")



import uuid
from threading import Lock
from typing import Optional

import pandas as pd
import pyarrow as pa

from odp.tabular_v2.big import BigCol
from odp.tabular_v2.big.big import MAX_BIGFILE_SIZE, STR_LIMIT, STR_MIN


def convert_schema_inward(schema: pa.Schema) -> pa.Schema:
    """add .ref fields for columns marked with big, helper only used by create()"""
    return Buffer(None).with_outer_schema(schema).inner_schema


class Buffer:
    def __init__(self, parent: Optional[BigCol]):
        self.data = []
        self.size = 0
        self.next_id = uuid.uuid4().hex
        self.lock = Lock()
        self.parent = parent
        self.big_fields = []
        self.small_fields = []
        self.inner_schema = None

    def with_inner_schema(self, inner_schema: pa.Schema):
        self.inner_schema = inner_schema
        for name in inner_schema.names:
            if name.endswith(".ref"):
                continue
            f = inner_schema.field(name)
            if f.type != pa.string() and f.type != pa.binary():
                continue
            meta = f.metadata
            if meta and b"big" in meta:
                self.big_fields.append(name)
            else:
                self.small_fields.append(name)
        return self

    def with_outer_schema(self, outer_schema: pa.Schema) -> "Buffer":
        fields = []
        for name in outer_schema.names:
            field: pa.Field = outer_schema.field(name)
            fields.append(field)
            if field.type != pa.string() and field.type != pa.binary():
                continue
            meta = field.metadata
            if meta and b"big" in meta:
                fields.append(pa.field(name + ".ref", pa.string()))
                self.big_fields.append(name)
            else:
                self.small_fields.append(name)
        self.inner_schema = pa.schema(fields)
        return self

    def encode(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        for name in self.big_fields:
            ref_name = name + ".ref"
            if ref_name not in df:
                df[ref_name] = None
        return df.apply(self.append, axis=1)

    def append(self, row):
        for name in self.big_fields:
            row[name + ".ref"] = None
            data = row[name]
            if data is None:
                continue
            if isinstance(data, str):
                data = data.encode("utf-8")  # convert to bytes
            size = len(data)  # size in bytes
            if size < STR_LIMIT:
                continue
            with self.lock:
                ref = f"{self.next_id}:{self.size}:{size}"  # ref to the current position
                self.data.append(data)  # append the new data to the buffer
                self.size += size  # update the size of the buffer
                if self.size > MAX_BIGFILE_SIZE:  # too much data? flush
                    self.parent.upload(self.next_id, self.data)
                    self.next_id = uuid.uuid4().hex
                    self.data = []
                    self.size = 0
            row[name + ".ref"] = ref
            row[name] = row[name][0:STR_MIN]
        return row

    def flush(self):
        with self.lock:
            if self.size > 0:
                self.parent.upload(self.next_id, self.data)
                self.data = []
                self.size = 0
                self.next_id = uuid.uuid4().hex
